import argparse
import sys
from datetime import date, datetime

from sdmcli import settings
from sdmcli.logging import log
from sdmcli.models import Category, Item, Ledger, db_conn


@db_conn
def ledger_add(args):
    Ledger.create(name=args.name)


@db_conn
def ledger_list(args):
    log.info("NAME")
    for ledger in Ledger.select().order_by(Ledger.name):
        log.info(ledger.name)


@db_conn
def ledger_update(args):
    ledger = Ledger.get(Ledger.name == args.name)
    ledger.name = args.new_name
    ledger.save()


@db_conn
def ledger_remove(args):
    ledger = Ledger.get(Ledger.name == args.name)
    ledger.delete_instance()


@db_conn
def category_add(args):
    ledger = Ledger.get(Ledger.name == args.ledger_name)
    Category.create(ledger=ledger, name=args.name)


@db_conn
def category_list(args):
    ledger = Ledger.get(Ledger.name == args.ledger_name)

    log.info("NAME")
    for category in ledger.categories.order_by(Category.name):
        log.info(category.name)


@db_conn
def category_update(args):
    category = Category.get_unique(args.ledger_name, args.name)
    category.name = args.new_name
    category.save()


@db_conn
def category_remove(args):
    category = Category.get_unique(args.ledger_name, args.name)
    category.delete_instance()


@db_conn
def item_add(args):
    ledger = Ledger.get(Ledger.name == args.ledger_name)

    category = None
    if args.category_name:
        category = Category.get_unique(ledger.name, args.category_name)

    Item.create(
        ledger=ledger,
        date=args.date,
        type=args.type,
        description=args.description,
        value=args.value,
        category=category,
    )


def str_to_date(value):
    return datetime.strptime(value, "%Y-%m-%d").date()


@db_conn
def item_list(args):
    ledger = Ledger.get(Ledger.name == args.ledger_name)

    today = date.today()
    from_date = str_to_date(args.from_date) if args.from_date else today.replace(day=1)
    to_date = str_to_date(args.to_date) if args.to_date else today
    items = ledger.items.where((Item.date >= from_date) & (Item.date <= to_date))

    if args.type:
        items = items.where(Item.type == args.type)

    if args.category_name:
        category = Category.get(Category.name == args.category_name)
        items = items.where(Item.category == category)

    template = "{id:<6}|{date:12}|{description:64}|{expense:<10}|{income:<10}|{category:32}"
    log.info(
        template.format(
            id="ID",
            date="DATE",
            description="DESCRIPTION",
            expense="EXPENSE",
            income="INCOME",
            category="CATEGORY",
        )
    )
    for item in items.order_by(Item.date):
        log.info(
            template.format(
                id=item.id,
                date=str(item.date),
                description=item.description,
                expense=item.value if item.type == "expense" else "",
                income=item.value if item.type == "income" else "",
                category=item.category.name if item.category else "",
            )
        )


@db_conn
def item_update(args):
    item = Item.get_by_id(args.id)

    item.date = args.date if args.date is not None else item.date
    item.type = args.type if args.type is not None else item.type
    item.description = args.description if args.description is not None else item.description
    item.value = args.value if args.value is not None else item.value

    if args.category_name:
        item.category = Category.get_unique(item.ledger.name, args.category_name)
    elif args.category_name == "":
        item.category = None

    item.save()


@db_conn
def item_remove(args):
    item = Item.get_by_id(args.id)
    item.delete_instance()


def currency(value):
    try:
        value = float(value)
    except ValueError:
        raise argparse.ArgumentTypeError("not a number")

    if value < 0:
        raise argparse.ArgumentTypeError("negative number")

    if len(str(value).split(".")[1]) > 2:
        raise argparse.ArgumentTypeError("too many decimal places")

    return value


def get_parser():
    parser = argparse.ArgumentParser(
        prog=settings.PROG,
        description=settings.DESCRIPTION,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparsers = parser.add_subparsers(title="commands", required=True)

    parser_ledger = subparsers.add_parser("ledger", help="collections of items and categories")
    subparsers_ledger = parser_ledger.add_subparsers(title="commands", required=True)

    parser_ledger_add = subparsers_ledger.add_parser("add")
    parser_ledger_add.add_argument("name")
    parser_ledger_add.set_defaults(func=ledger_add)

    parser_ledger_list = subparsers_ledger.add_parser("list")
    parser_ledger_list.set_defaults(func=ledger_list)

    parser_ledger_update = subparsers_ledger.add_parser("update")
    parser_ledger_update.add_argument("name")
    parser_ledger_update.add_argument("--name", required=True, dest="new_name")
    parser_ledger_update.set_defaults(func=ledger_update)

    parser_ledger_remove = subparsers_ledger.add_parser("remove")
    parser_ledger_remove.add_argument("name")
    parser_ledger_remove.set_defaults(func=ledger_remove)

    parser_category = subparsers.add_parser(
        "category", help="item categories, unique in each ledger"
    )
    subparsers_category = parser_category.add_subparsers(title="commands", required=True)

    parser_category_add = subparsers_category.add_parser("add")
    parser_category_add.add_argument("ledger_name")
    parser_category_add.add_argument("name")
    parser_category_add.set_defaults(func=category_add)

    parser_category_list = subparsers_category.add_parser("list")
    parser_category_list.add_argument("ledger_name")
    parser_category_list.set_defaults(func=category_list)

    parser_category_update = subparsers_category.add_parser("update")
    parser_category_update.add_argument("ledger_name")
    parser_category_update.add_argument("name")
    parser_category_update.add_argument("--name", required=True, dest="new_name")
    parser_category_update.set_defaults(func=category_update)

    parser_category_remove = subparsers_category.add_parser("remove")
    parser_category_remove.add_argument("ledger_name")
    parser_category_remove.add_argument("name")
    parser_category_remove.set_defaults(func=category_remove)

    parser_item = subparsers.add_parser("item", help="expenses and incomes")
    subparsers_item = parser_item.add_subparsers(title="commands", required=True)

    parser_item_add = subparsers_item.add_parser("add")
    parser_item_add.add_argument("ledger_name")
    parser_item_add.add_argument("date")
    parser_item_add.add_argument("type", choices=["expense", "income"])
    parser_item_add.add_argument("description")
    parser_item_add.add_argument("value", type=currency)
    parser_item_add.add_argument("--category", dest="category_name")
    parser_item_add.set_defaults(func=item_add)

    parser_item_list = subparsers_item.add_parser("list")
    parser_item_list.add_argument("ledger_name")
    parser_item_list.add_argument("--from", dest="from_date")
    parser_item_list.add_argument("--to", dest="to_date")
    parser_item_list.add_argument("--type", choices=["expense", "income"])
    parser_item_list.add_argument("--category", dest="category_name")
    parser_item_list.set_defaults(func=item_list)

    parser_item_update = subparsers_item.add_parser("update")
    parser_item_update.add_argument("id")
    parser_item_update.add_argument("--date")
    parser_item_update.add_argument("--type", choices=["expense", "income"])
    parser_item_update.add_argument("--description")
    parser_item_update.add_argument("--value", type=currency)
    parser_item_update.add_argument("--category", dest="category_name")
    parser_item_update.set_defaults(func=item_update)

    parser_item_remove = subparsers_item.add_parser("remove")
    parser_item_remove.add_argument("id")
    parser_item_remove.set_defaults(func=item_remove)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser
